import datetime
import pytz

from aio_client.provider.api import PostProviderRequest, push_request
from django.db.models import Prefetch
from kinder.core.declaration.models import Declaration, DeclarationUnit
from kinder.core.queue_module.api import get_info_for_declaration
from kinder.core.queue_module.enum import COMMON_QUEUE_STATUSES
from kinder.settings import TIME_ZONE

from concentrator.smev3_v321.base.constants import SUCCESS_MESSAGE
from concentrator.smev3_v321.base.utils import (BaseExecutor, SMEV3Response,
                                                get_oktmo_region,
                                                is_cancel_allowed,
                                                render_type2xml)
from concentrator.smev3_v321.service_types import kinder_conc

from .constants import GetApplicationQueueMessage
from ..helpers import get_declaration_by_client_or_portal_id


class GetApplicationQueueRequestExecutor(BaseExecutor):
    """Исполнитель сервиса GetApplicationQueueRequest."""

    name_service = 'GetApplicationQueueRequest'
    type_service = kinder_conc.GetApplicationQueueRequestType

    @classmethod
    def prepare_query(cls):
        """Выполняет подготовку запроса."""

        subquery_declaration_unit_ids = (
            DeclarationUnit.objects.only('unit_id'))

        query = Declaration.objects.prefetch_related(
            Prefetch('declarationunit_set',
                     queryset=subquery_declaration_unit_ids,
                     to_attr='declaration_unit_ids'))

        return query

    @classmethod
    def process(cls, message, request_body):

        order_id = request_body.GetApplicationQueueRequest.orderId

        response_body = render_type2xml(
            cls.get_response(order_id), name_type='FormDataResponse')

        response = push_request(PostProviderRequest(
            origin_message_id=message['origin_message_id'],
            body=response_body,
            message_type=message['message_type'],
            replay_to=message['replay_to'],
            is_test_message=message['is_test_message']))

        return SMEV3Response(
            response,
            {
                'method_name': cls.name_service,
                'response': response_body,
                'result': SUCCESS_MESSAGE
            })

    @classmethod
    def get_response(cls, order_id):
        """Формирует ответ на GetApplicationQueueRequest."""

        declaration, _ = get_declaration_by_client_or_portal_id(
            cls.prepare_query(), order_id)

        validation_error_response = cls.validate(declaration)
        if validation_error_response:
            return validation_error_response

        declaration_info = get_info_for_declaration(declaration, only_mo=True)
        if declaration_info:
            declaration_info = declaration_info[0]

        get_app_queue_type = (
            kinder_conc.ApplicationQueueResponseType(
                orderId=int(declaration.client_id),
                Position=declaration_info['position'],
                Total=declaration_info['total'],
                WithoutQueue=declaration_info['out_of_order'],
                FirstQueue=declaration_info['first_order'],
                AdvantageQueue=declaration_info['preemptive_right'],
                RelevantDT=datetime.datetime.now(
                    pytz.timezone(TIME_ZONE)).replace(microsecond=0)
            )
        )

        response = kinder_conc.FormDataResponseType(
            oktmo=get_oktmo_region(),
            GetApplicationQueueResponse=get_app_queue_type)

        return response

    @classmethod
    def validate(cls, declaration):
        """Валидирует заявление, если валидация не пройдена - возвращает ответ
         с ошибкой"""

        kwargs = {}

        # Проверка существования заявления
        if not declaration:
            kwargs.update({
                'statusCode': kinder_conc.statusCodeType(
                    GetApplicationQueueMessage.values[
                        GetApplicationQueueMessage.NOT_EXISTS]),
                'comment': GetApplicationQueueMessage.NOT_EXISTS})

        # Заявка должна быть со статусом, участвующим в очереди
        elif declaration.status.code not in COMMON_QUEUE_STATUSES:
            kwargs.update({
                'statusCode': kinder_conc.statusCodeType(
                    GetApplicationQueueMessage.values[
                        GetApplicationQueueMessage.NOT_QUEUED]),
                'comment': GetApplicationQueueMessage.NOT_QUEUED,
                'cancelAllowed': is_cancel_allowed(declaration.status)
            })

        if kwargs:
            kwargs.update({
                'orderId': kinder_conc.orderIdType(int(declaration.client_id)),
            })

            return kinder_conc.FormDataResponseType(
                changeOrderInfo=kinder_conc.changeOrderInfoType(**kwargs))
